package evgeniy.martian.base;

import evgeniy.martian.base.app.Application;
import evgeniy.martian.base.app.ConsoleApplication;
import evgeniy.martian.base.app.LevelRenderer;
import evgeniy.martian.base.app.UiApplication;
import evgeniy.martian.base.level.Level;
import evgeniy.martian.base.level.LevelParser;
import evgeniy.martian.base.objects.Manager;
import evgeniy.martian.base.objects.MoveTaskManager;
import evgeniy.martian.base.utils.Position;
import evgeniy.martian.base.utils.Utils;

import javax.swing.*;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.Arrays;
import java.util.Deque;
import java.util.List;

public class Main {

    public static void main(String[] args) {
        String workingDirectory = System.getProperty("user.dir");

        Path workingDirectoryPath = Paths.get(workingDirectory);
        Path outputFilePath = workingDirectoryPath.resolve("moves.txt");
        Path mapsDirectoryPath = workingDirectoryPath.resolve("maps");

        Path levelFilePath = mapsDirectoryPath.resolve("mapFile.txt");
        boolean console = false;

        for (String argument : args) {
            if (argument.equals("-console")) {
                console = true;
            }
            else if (argument.endsWith(".txt")) {
                Path levelFilePathArgument = mapsDirectoryPath.resolve(argument);
                if (Files.exists(levelFilePathArgument) && !Files.isDirectory(levelFilePathArgument)) {
                    levelFilePath = levelFilePathArgument;
                }
            }
        }

        if (Files.notExists(levelFilePath)) {
            System.out.printf("Файл карты '%s' не нейден%n", levelFilePath);
            return;
        }
        if (Files.isDirectory(levelFilePath)) {
            System.out.printf("Файл карты '%s' папка, ожидался файл%n", levelFilePath);
            return;
        }

        LevelParser parser = new LevelParser(levelFilePath);
        Level level = parser.parse();

        if (level == null) {
            if (console) System.out.println("Ошибка при загрузке карты");
            else JOptionPane.showMessageDialog(null, "Ошибка при загрузке карты", "Ошибка загрузки", JOptionPane.ERROR_MESSAGE);
            return;
        }

        Manager manager = new Manager();
        manager.setLevel(level);

        manager.getTaskManager().initializeMap();

        Application application;
        if (console) {
            application = new ConsoleApplication();
        }
        else {
            application = new UiApplication(manager);

            ((UiApplication) application).getRenderer().getCameraOffset().setFrom(new Position(
                    UiApplication.MIN_SIZE.width / 2 - level.getWidth() * LevelRenderer.TILE_SIZE / 2,
                    UiApplication.MIN_SIZE.height / 2 - level.getHeight() * LevelRenderer.TILE_SIZE / 2
            ));

            new Thread(((UiApplication) application)::launch).start();
        }

        long updatePeriod = 200;

        boolean enabled = true;

        MoveTaskManager.Status status = MoveTaskManager.Status.SUCCESS;
        while (enabled) {
            long start = System.currentTimeMillis();

            status = manager.getTaskManager().step();
            application.updateUi();

            if (status != MoveTaskManager.Status.SUCCESS) {
                enabled = false;
            }

            long end = System.currentTimeMillis();

            if (!console) {
                long sleep = updatePeriod - (end - start);
                if (sleep < 1) continue;

                try {
                    Thread.sleep(sleep);
                }
                catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
        }

        switch (status) {
            case FINISHED:
                application.completeDialog();
                break;
            case FAILURE:
                application.failureDialog();
                break;
            case DEAD:
                application.playerDeadDialog();
                break;
            case PATH_NOT_FOUND:
                application.pathNotFoundDialog();
                break;
        }

        if (status == MoveTaskManager.Status.FINISHED) {
            Deque<Position> fullPath = manager.getTaskManager().getFullPath();
            if (!fullPath.isEmpty()) {
                StringBuilder builder = new StringBuilder();

                Position first = fullPath.removeFirst();
                while (!fullPath.isEmpty()) {
                    Position second = fullPath.removeFirst();
                    builder.append(Utils.toDirectionSymbol(first, second));
                    first = second;
                }

                List<String> lines = Arrays.asList(String.valueOf(builder.length()), builder.toString());
                try {
                    Utils.write(outputFilePath, lines, StandardCharsets.UTF_8, StandardOpenOption.CREATE, StandardOpenOption.WRITE);
                }
                catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }

    }

}
