package evgeniy.martian.base.level;

import evgeniy.martian.base.level.tile.*;
import evgeniy.martian.base.utils.Position;

import java.util.ArrayList;
import java.util.List;

public class Level {

    private final int width;
    private final int height;

    private final Tile[][] level;
    private final Tile[][] cubes;

    private Position startPosition;
    private Position endPosition;

    public Level(int height, int width) {
        this.height = height;
        this.width = width;

        this.level = new Tile[height][width];
        this.cubes = new Tile[height][width];
    }

    public int getWidth() {
        return this.width;
    }

    public int getHeight() {
        return this.height;
    }

    public void set(int x, int y, Tile tile) {
        level[x][y] = tile;
    }

    public void setCube(int x, int y, CubeTile tile) {
        cubes[x][y] = tile;
    }

    public Tile get(int x, int y) {
        return level[x][y];
    }

    public CubeTile getCube(int x, int y) {
        return (CubeTile) cubes[x][y];
    }

    public Tile get(Position position) {
        return get(position.getX(), position.getY());
    }

    public CubeTile getCube(Position position) {
        return getCube(position.getX(), position.getY());
    }

    public Position getStartPosition() {
        return this.startPosition;
    }

    public void setStartPosition(Position position) {
        this.startPosition = position;
    }

    public Position getEndPosition() {
        return this.endPosition;
    }

    public void setEndPosition(Position endPosition) {
        this.endPosition = endPosition;
    }

    public List<Position> getDoors() {
        List<Position> result = new ArrayList<>();
        for (int x = 0; x < height; x++) {
            for (int y = 0; y < width; y++) {
                Tile tile = get(x, y);
                if (tile.is(Tile.Type.DOOR)) {
                    result.add(new Position(x, y));
                }
            }
        }
        return result;
    }

    public List<Position> getButtons() {
        List<Position> result = new ArrayList<>();
        for (int x = 0; x < height; x++) {
            for (int y = 0; y < width; y++) {
                Tile tile = get(x, y);
                if (tile.is(Tile.Type.BUTTON)) {
                    result.add(new Position(x, y));
                }
            }
        }
        return result;
    }

    public List<Position> getCubes() {
        List<Position> result = new ArrayList<>();
        for (int x = 0; x < height; x++) {
            for (int y = 0; y < width; y++) {
                Tile tile = getCube(x, y);
                if (tile != null) result.add(new Position(x, y));
            }
        }
        return result;
    }

    public boolean buttonIsPressed(ButtonTile.Variant variant) {
        for (int x = 0; x < height; x++) {
            for (int y = 0; y < width; y++) {
                Tile tile = get(x, y);
                if (tile.is(Tile.Type.BUTTON)) {
                    ButtonTile buttonTile = (ButtonTile) tile;
                    ButtonTile.Variant currVariant = buttonTile.getVariant();
                    if (currVariant == variant) {
                        return getCube(x, y) != null;
                    }
                }
            }
        }
        return false;
    }

    public void dump() {
        StringBuilder builder = new StringBuilder();
        for (int x = 0; x < this.level.length; x++) {
            Tile[] xTiles = this.level[x];
            for (int y = 0; y < xTiles.length; y++) {
                Tile tile = xTiles[y];
                if (tile == null) builder.append("?");
                else builder.append(tile.name());
            }
            builder.append("\n");
        }
        System.out.println(builder);
    }

    public Position getButtonFor(Position doorPosition) {
        Tile tile = get(doorPosition);
        if (tile.is(Tile.Type.DOOR)) {
            DoorTile door = (DoorTile) tile;
            ButtonTile.Variant button = door.getVariant().getButton();

            for (Position position : getButtons()) {
                ButtonTile buttonTile = (ButtonTile) get(position);
                if (buttonTile.getVariant() == button) return position;
            }
        }

        return null;
    }

    public List<Position> getPortals() {
        List<Position> result = new ArrayList<>();
        for (int x = 0; x < height; x++) {
            for (int y = 0; y < width; y++) {
                Tile tile = get(x, y);
                if (tile != null && (tile.is(Tile.Type.BLUE_PORTAL) || tile.is(Tile.Type.ORANGE_PORTAL))) {
                    result.add(new Position(x, y));
                }
            }
        }
        return result;
    }

    public Position getPortal(OrangePortalTile.Variant variant) {
        for (Position portal : getPortals()) {
            Tile tile = get(portal);
            if (tile.is(Tile.Type.ORANGE_PORTAL)) {
                OrangePortalTile orangePortal = (OrangePortalTile) tile;
                if (variant == orangePortal.getVariant()) {
                    return portal;
                }
            }
        }
        return null;
    }

    public Position getPortal(BluePortalTile.Variant variant) {
        for (Position portal : getPortals()) {
            Tile tile = get(portal);
            if (tile.is(Tile.Type.BLUE_PORTAL)) {
                BluePortalTile bluePortal = (BluePortalTile) tile;
                if (variant == bluePortal.getVariant()) {
                    return portal;
                }
            }
        }
        return null;
    }


    public enum Moves {

        RIGHT, LEFT, UP, DOWN

    }

}
