package evgeniy.martian.base.level;

import evgeniy.martian.base.level.tile.*;
import evgeniy.martian.base.utils.Position;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Collections;
import java.util.List;

public class LevelParser {

    private final Path levelPath;

    public LevelParser(Path levelPath) {
        this.levelPath = levelPath;
    }

    public Level parse() {
        if (Files.notExists(levelPath)) return null;

        List<String> levelLines;

        try {
            levelLines = Files.readAllLines(this.levelPath, StandardCharsets.UTF_8);
        }
        catch (IOException e) {
            e.printStackTrace();
            levelLines = Collections.emptyList();
        }

        if (levelLines.isEmpty()) return null;
        String levelProperties = levelLines.remove(0);

        String[] levelSize = levelProperties.split("\\s+");
        if (levelSize.length != 2) return null;

        int sizeH = -1;
        int sizeW = -1;

        try {
            sizeH = Integer.parseInt(levelSize[0]);
            sizeW = Integer.parseInt(levelSize[1]);
        }
        catch (NumberFormatException e) {
            e.printStackTrace();
        }

        if (sizeH < 10 || sizeH > 50) return null;
        if (sizeW < 10 || sizeW > 50) return null;

        if (levelLines.size() < sizeH) return null;
        String[][] inputLevel = new String[sizeH][sizeW];

        for (int index = 0; index < sizeH; index++) {
            String levelInfo = levelLines.get(index);
            String[] levelObjects = levelInfo.split("\\s+");
            if (levelObjects.length < sizeW) return null;

            System.arraycopy(levelObjects, 0, inputLevel[index], 0, sizeW);
        }

        return deserializeLevel(sizeH, sizeW, inputLevel);
    }

    private Level deserializeLevel(int height, int width, String[][] input) {
        /*
            . - floor
            X - wall
            S - start
            Q - exit

            K - cube

            A, B, C, D - door
            a, b, c, d - button

            F, G, H, I, J - blue portals
            f, g, h, i, j - orange portals
         */

        Level level = new Level(height, width);
        for (int x = 0; x < height; x++) {
            for (int y = 0; y < width; y++) {
                String id = input[x][y];
                switch (id) {
                    case ".":
                        level.set(x, y, new FloorTile(x, y));
                        break;
                    case "X":
                        level.set(x, y, new WallTile(x, y));
                        break;
                    case "S":
                        level.set(x, y, new StartTile(x, y));
                        level.setStartPosition(new Position(x, y));
                        break;
                    case "Q":
                        level.set(x, y, new ExitTile(x, y));
                        level.setEndPosition(new Position(x, y));
                        break;
                    case "K":
                        level.set(x, y, new FloorTile(x, y));
                        level.setCube(x, y, new CubeTile(x, y));
                        break;
                    case "A":
                    case "B":
                    case "C":
                    case "D":
                        level.set(x, y, new DoorTile(x, y, DoorTile.Variant.from(id)));
                        break;
                    case "a":
                    case "b":
                    case "c":
                    case "d":
                        level.set(x, y, new ButtonTile(x, y, ButtonTile.Variant.from(id)));
                        break;
                    case "F":
                    case "G":
                    case "H":
                    case "I":
                    case "J":
                        level.set(x, y, new BluePortalTile(x, y, BluePortalTile.Variant.from(id)));
                        break;
                    case "f":
                    case "g":
                    case "h":
                    case "i":
                    case "j":
                        level.set(x, y, new OrangePortalTile(x, y, OrangePortalTile.Variant.from(id)));
                        break;

                }
            }
        }
        return level;
    }

}
