package evgeniy.martian.base.objects;

import evgeniy.martian.base.level.Level;
import evgeniy.martian.base.level.tile.*;
import evgeniy.martian.base.utils.ButtonsSolver;
import evgeniy.martian.base.utils.CubeUtils;
import evgeniy.martian.base.utils.Position;

import java.util.*;

import static evgeniy.martian.base.objects.pathfinding.AStarPathfinding.aStar;

public class MoveTaskManager2 {

    private final Manager manager;

    private Deque<Goal> goals = new ArrayDeque<>();
    private Deque<Position> currentPath;
    private Level level;

    private Deque<Position> fullPath = new ArrayDeque<>();

    public MoveTaskManager2(Manager manager) {
        this.manager = manager;
    }

    public Deque<Position> getFullPath() {
        return this.fullPath;
    }

    public void initializeMap() {
        level = manager.getLevel();
        goals.addFirst(new Goal(level.getEndPosition(), true));

        fullPath.addFirst(level.getStartPosition());
    }

    public Status step() {

        if (currentPath == null || currentPath.isEmpty()) {

            if (goals.isEmpty()) {
                if (manager.getPlayerPosition().equals(level.getEndPosition().toMutable())) {
                    return Status.FINISHED;
                }
                else {
                    return Status.FAILURE;
                }
            }

            return nextGoal();
        }

        Position newPosition = currentPath.removeFirst();
        Position oldPosition = manager.getPlayerPosition().toImmutable();

        boolean changePlayerPosition = true;

        CubeTile cube = level.getCube(newPosition);
        if (cube != null) {
            Position offset = newPosition.sub(oldPosition);
            Position nextCubePostion = newPosition.plus(offset);
            if (level.getCube(nextCubePostion) != null) changePlayerPosition = false;
            Tile tile = level.get(nextCubePostion);

            if (tile == null) changePlayerPosition = false;
            else if (!tile.is(Tile.Type.FLOOR) && !tile.is(Tile.Type.BUTTON)) {
                System.out.println(tile.getType());
                changePlayerPosition = false;
            }

            if (changePlayerPosition) {
                level.setCube(cube.getX(), cube.getY(), null);
                level.setCube(nextCubePostion.getX(), nextCubePostion.getY(), new CubeTile(nextCubePostion.getX(), nextCubePostion.getY()));
            }
        }

        if (changePlayerPosition) {
            manager.getPlayerPosition().setFrom(newPosition);
            fullPath.addLast(newPosition);

            if (!oldPosition.equals(newPosition)) {
                Tile tile = level.get(newPosition);
                if (tile != null && tile.is(Tile.Type.ORANGE_PORTAL)) {
                    OrangePortalTile orangePortal = (OrangePortalTile) tile;
                    Position portal = level.getPortal(orangePortal.getVariant().toBlue());
                    if (portal == null) throw new UnsupportedOperationException();
                    manager.getPlayerPosition().setFrom(portal);
                }
                else if (tile != null && tile.is(Tile.Type.BLUE_PORTAL)) {
                    BluePortalTile bluePortal = (BluePortalTile) tile;
                    Position portal = level.getPortal(bluePortal.getVariant().toOrange());
                    if (portal == null) throw new UnsupportedOperationException();
                    manager.getPlayerPosition().setFrom(portal);
                }
            }
        }

        if (newPosition.equals(goals.getFirst().target)) {
            goals.removeFirst();
            currentPath.clear();
        }

        if (newPosition.equals(oldPosition)) {
            return Status.REVOKE;
        }

        return Status.SUCCESS;
    }

    public Status nextGoal() {
        Goal goal = goals.getFirst();

        Position currentPos = manager.getPlayerPosition().toImmutable();
        Deque<Position> path = aStar(manager, currentPos, goal.target);

        if (path.isEmpty()) return Status.PATH_NOT_FOUND;

        Position prev = manager.getPlayerPosition().toImmutable();
        for (Position position : path) {
            Tile tile = level.get(position);

            if (tile.is(Tile.Type.DOOR)) {
                if (!level.buttonIsPressed(((DoorTile) tile).getVariant().getButton())) {
                    return handleDoor(position);
                }
            }

            if (goal.viewCubes) {
                CubeTile cube = level.getCube(position);
                if (cube != null) {
                    return handleCube(prev, position);
                }
            }

        }

        this.currentPath = path;
        return Status.REVOKE;
    }

    private Status handleCube(Position currentPosition, Position cube) {
        List<CubeUtils.Move> moves = CubeUtils.availableCubeMoves(manager, cube, currentPosition);
        if (moves.isEmpty()) throw new UnsupportedOperationException();

        if (moves.size() == 1) {
            goals.addFirst(new Goal(cube, false));
            this.currentPath = null;
            return Status.REVOKE;
        }

        moveFor: for (CubeUtils.Move move : moves) {
            Position nextCubePosition = move.plus(cube);
            Position nextPlayerPosition = cube;

            Deque<Position> positions = aStar(manager, nextPlayerPosition, goals.getFirst().target);
            for (Position position : positions) {
                if (position.equals(nextCubePosition)) continue moveFor;
            }
            goals.addFirst(new Goal(nextPlayerPosition, false));
            goals.addFirst(new Goal(move.sub(cube), false));
            return Status.REVOKE;
        }

        return Status.FAILURE;
    }

    private Status handleDoor(Position doorPosition) {
        Position playerPosition = manager.getPlayerPosition().toImmutable();
        Position buttonPosition = level.getButtonFor(doorPosition);

        Deque<Position> path = ButtonsSolver.solve(manager, playerPosition, buttonPosition);

        if (path.size() == 1) {
            System.out.println("F1");
            return Status.FAILURE;
        }

        boolean first = true;

        Position last = path.removeLast();
        while (!path.isEmpty()) {
            Position position = path.removeLast();
            Position off = last.sub(position);

            if (first || path.isEmpty()) {
                goals.addFirst(new Goal(position, false));
                first = false;
            }
            goals.addFirst(new Goal(position.sub(off), false));
            last = position;
        }
        currentPath = null;

        return Status.REVOKE;
    }

    public enum Status {

        FINISHED,
        FAILURE,
        DEAD,
        PATH_NOT_FOUND,
        SUCCESS,
        REVOKE

    }

    public static class Goal {

        Position target;
        boolean viewCubes = false;

        public Goal(Position target, boolean viewCubes) {
            this.target = target;
            this.viewCubes = viewCubes;
        }

        @Override
        public String toString() {
            return "Goal{%s}".formatted(target);
        }
    }

}
