package evgeniy.martian.base.objects.pathfinding;

import evgeniy.martian.base.level.Level;
import evgeniy.martian.base.level.tile.BluePortalTile;
import evgeniy.martian.base.level.tile.OrangePortalTile;
import evgeniy.martian.base.level.tile.Tile;
import evgeniy.martian.base.objects.Manager;
import evgeniy.martian.base.utils.Position;

import java.util.*;

import static java.lang.Math.abs;

public class AStarPathfindingPortals {

    private final Position start;
    private final Position end;
    private final Level level;

    private final int width;
    private final int height;

    private Entry[][] field;
    private boolean[][] closed;
    private PriorityQueue<Entry> opened;

    public AStarPathfindingPortals(Position start, Position end, Level level) {
        this.start = start;
        this.end = end;
        this.level = level;

        this.width = level.getWidth();
        this.height = level.getHeight();
    }

    public Deque<Position> getPath(Manager manager) {
        init(manager);
        process(manager);
        return buildPath();
    }

    private void init(Manager manager) {
        field = new Entry[height][width];
        closed = new boolean[height][width];

        opened = new PriorityQueue<>(Comparator.comparingInt(f -> f.fCost));

        for (int x = 0; x < height; x++) {
            for (int y = 0; y < width; y++) {
                Entry entry = new Entry(new Position(x, y));
                entry.hCost = abs(end.getX() - entry.position.getX()) + abs(end.getY() - entry.position.getY());
                entry.canPass = level.get(x, y).canPass(manager);
                field[x][y] = entry;
            }
        }

        opened.add(field[start.getX()][start.getY()]);
    }

    private void process(Manager manager) {
        while (!opened.isEmpty()) {
            Entry current = opened.poll();
            Position position = current.position;

            closed[position.getX()][position.getY()] = true;
            if (position.equals(end)) {
                return;
            }

            List<Entry> neighbors = getNeighbors(current);
            for (Entry neighbor : neighbors) {

                updateEntry(current, neighbor);
            }
        }
    }

    private Deque<Position> buildPath() {
        Deque<Position> path = new ArrayDeque<>();

        Entry current = field[this.end.getX()][this.end.getY()];
        if (current.parent == null) return path;

        while(current != null) {
            path.addFirst(current.position);
            current = current.parent;
        }

        return path;
    }

    private void updateEntry(Entry current, Entry target) {
        Position position = target.position;
        if (closed[position.getX()][position.getY()] || !target.canPass) return;

        int fCost = target.hCost + current.fCost + Entry.COST;
        boolean isOpen = opened.contains(target);

        if (!isOpen || fCost < target.fCost) {
            target.fCost = fCost;
            target.parent = current;

            if (!isOpen) {
                opened.add(target);
            }
        }
    }

    private List<Entry> getNeighbors(Entry entry) {
        Position position = entry.position;
        int x = position.getX();
        int y = position.getY();

        Tile tile = level.get(position);
        if (tile != null && tile.is(Tile.Type.ORANGE_PORTAL)) {
            OrangePortalTile orangePortal = (OrangePortalTile) tile;
            Position portal = level.getPortal(orangePortal.getVariant().toBlue());
            if (portal != null) {
                x = portal.getX();
                y = portal.getY();

                closed[x][y] = true;
            }
        }
        else if (tile != null && tile.is(Tile.Type.BLUE_PORTAL)) {
            BluePortalTile bluePortal = (BluePortalTile) tile;
            Position portal = level.getPortal(bluePortal.getVariant().toOrange());
            if (portal != null) {
                x = portal.getX();
                y = portal.getY();

                closed[x][y] = true;
            }
        }

        List<Entry> result = new ArrayList<>();

        if (x - 1 > -1) result.add(field[x - 1][y]);
        if (x + 1 < height) result.add(field[x + 1][y]);
        if (y - 1 > -1) result.add(field[x][y - 1]);
        if (y + 1 < width) result.add(field[x][y + 1]);
        return result;
    }

    public static Deque<Position> aStar(Manager manager, Position start, Position end) {
        AStarPathfindingPortals pathfinding = new AStarPathfindingPortals(start, end, manager.getLevel());
        return pathfinding.getPath(manager);
    }

    public static class Entry {

        public static final int COST = 10;

        private Position position;
        private Entry parent;

        private int hCost;
        private int fCost;

        private boolean canPass = true;

        public Entry(Position position) {
            this.position = position;
        }

    }

}
